"""summarize metagenome gather results"""

usage="""

    sourmash tax metagenome --gather-csv <gather_csv> [ ... ] --taxonomy-csv <taxonomy-csv> [ ... ]

The 'tax metagenome' command reads in metagenome gather result CSVs and
summarizes by taxonomic lineage.

The default output format consists of four columns,
 'query_name,rank,fraction,lineage', where 'fraction' is the fraction
 of the query matched to that reported rank and lineage. The summarization
 is reported for each taxonomic rank.

Alternatively, you can output results at a specific rank (e.g. species)
in 'krona', 'lineage_summary', and 'human' formats.

Use '-F human' to display human-readable output.

Please see the 'tax metagenome' documentation for more details:
  https://sourmash.readthedocs.io/en/latest/command-line.html#sourmash-tax-metagenome-summarize-metagenome-content-from-gather-results
"""

import sourmash
from sourmash.logging import notify, print_results, error
from sourmash.cli.utils import add_rank_arg, check_rank, check_tax_outputs



def subparser(subparsers):
    subparser = subparsers.add_parser('metagenome',
                                      usage=usage)
    subparser.add_argument(
        '-g', '--gather-csv', action="extend", nargs='*', default = [],
        help='CSVs from sourmash gather'
    )
    subparser.add_argument(
        '--from-file',  metavar='FILE', default = None,
        help='input many gather results as a text file, with one gather CSV per line'
    )
    subparser.add_argument(
        '-q', '--quiet', action='store_true',
        help='suppress non-error output'
    )
    subparser.add_argument(
        '-o', '--output-base', default='-',
        help='base filepath for output file(s) (default stdout)'
    )
    subparser.add_argument(
        '--output-dir', default= "",
        help='directory for output files'
    )
    subparser.add_argument(
        '-t', '--taxonomy-csv', '--taxonomy', metavar='FILE',
        action="extend", nargs='+', required=True,
        help='database lineages CSV'
    )
    subparser.add_argument(
        '--keep-full-identifiers', action='store_true',
        help='do not split identifiers on whitespace'
    )
    subparser.add_argument(
        '--keep-identifier-versions', action='store_true',
        help='after splitting identifiers, do not remove accession versions'
    )
    subparser.add_argument(
        '--fail-on-missing-taxonomy', action='store_true',
        help='fail quickly if taxonomy is not available for an identifier',
    )
    subparser.add_argument(
        '-F', '--output-format', default=[], nargs='*', action="extend",
        choices=["human", "csv_summary", "krona", "lineage_summary", "kreport", "lingroup", "bioboxes"],
        help='choose output format(s)',
    )
    subparser.add_argument(
        '-f', '--force', action = 'store_true',
        help='continue past errors in taxonomy database loading',
    )
    subparser.add_argument(
        '--lins', '--lin-taxonomy', action='store_true', default=False,
        help="use LIN taxonomy in place of standard taxonomic ranks.  Note that the taxonomy CSV must contain 'lin' lineage information."
    )
    subparser.add_argument(
        '--lingroup', '--lingroups', metavar='FILE', default=None,
        help="CSV containing 'name', 'lin' columns, where 'lin' is the lingroup prefix. Will produce a 'lingroup' report containing taxonomic summarization for each group."
    )
    add_rank_arg(subparser)

def main(args):
    import sourmash
    try:
        if not args.gather_csv and not args.from_file:
            raise ValueError(f"No gather CSVs found! Please input via '-g' or '--from-file'.")
        if args.rank:
            args.rank = check_rank(args)
        args.output_format = check_tax_outputs(args, rank_required = ['krona', 'lineage_summary'], incompatible_with_lins = ['bioboxes', 'kreport'], use_lingroup_format=True)

    except ValueError as exc:
        error(f"ERROR: {str(exc)}")
        import sys; sys.exit(-1)

    return sourmash.tax.__main__.metagenome(args)



"""
Tests for the 'sourmash signature grep' command line.
"""
import shutil
import os
import csv
import gzip

import pytest

import sourmash_tst_utils as utils
import sourmash
from sourmash_tst_utils import SourmashCommandFailed
from sourmash.signature import load_signatures

## command line tests


def test_grep_1_sig_name(runtmp):
    # search on substring in name
    sig47 = utils.get_test_data('47.fa.sig')

    runtmp.run_sourmash('sig', 'grep', 'Shewanella', sig47)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella' in ss.name
    assert ss.md5sum() == '09a08691ce52952152f0e866a59f6261'


def test_grep_1_sig_name_case_sensitive(runtmp):
    # search on substring in name
    sig47 = utils.get_test_data('47.fa.sig')

    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash('sig', 'grep', 'shewanella', sig47)


def test_grep_1_sig_name_case_insensitive(runtmp):
    # search on substring in name, case insensitive
    sig47 = utils.get_test_data('47.fa.sig')

    runtmp.run_sourmash('sig', 'grep', '-i', 'shewanella', sig47)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella' in ss.name
    assert ss.md5sum() == '09a08691ce52952152f0e866a59f6261'


def test_grep_1_sig_name_exclude(runtmp):
    # search on substring in name, case insensitive
    sig47 = utils.get_test_data('47.fa.sig')

    # no matches!
    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash('sig', 'grep', '-v', 'Shewanella', sig47)


def test_grep_2_sig_md5(runtmp):
    # search on substring in md5
    sig47 = utils.get_test_data('47.fa.sig')

    runtmp.run_sourmash('sig', 'grep', 'ce52952152f0', sig47)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert ss.md5sum() == '09a08691ce52952152f0e866a59f6261'


def test_grep_2_sig_md5_case_sensitive(runtmp):
    # case sensitive no match
    sig47 = utils.get_test_data('47.fa.sig')

    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash('sig', 'grep', 'CE52952152f0', sig47)


def test_grep_2_sig_md5_case_insensitive(runtmp):
    # search on substring in md5, case insensitive
    sig47 = utils.get_test_data('47.fa.sig')

    runtmp.run_sourmash('sig', 'grep', '-i', 'CE52952152f0', sig47)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert ss.md5sum() == '09a08691ce52952152f0e866a59f6261'


def test_grep_3_filename(runtmp):
    # filename match
    sig47 = utils.get_test_data('47.fa.sig')

    runtmp.run_sourmash('sig', 'grep', '47.fa', sig47)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert '47.fa' in ss.filename
    assert ss.md5sum() == '09a08691ce52952152f0e866a59f6261'


def test_grep_3_filename_regexp(runtmp):
    # search for a regexp on filename
    sig47 = utils.get_test_data('47.fa.sig')

    runtmp.run_sourmash('sig', 'grep', '^47.fa', sig47)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert '7.fa' in ss.filename
    assert ss.md5sum() == '09a08691ce52952152f0e866a59f6261'


def test_grep_4_no_manifest(runtmp):
    # fail search when no manifest, by default
    sbt = utils.get_test_data('v6.sbt.zip')

    with pytest.raises(SourmashCommandFailed) as exc:
        runtmp.run_sourmash('sig', 'grep', 'e60265', sbt)

    print(runtmp.last_result.err)
    assert 'ERROR on filename' in runtmp.last_result.err
    assert 'sig grep requires a manifest by default, but no manifest present.' in runtmp.last_result.err


def test_grep_4_no_manifest_ok(runtmp):
    # generate manifest if --no-require-manifest
    sbt = utils.get_test_data('v6.sbt.zip')

    runtmp.run_sourmash('sig', 'grep', 'e60265', sbt, '--no-require-manifest')

    ss = load_signatures(runtmp.last_result.out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'e60265' in ss.md5sum()


def test_grep_5_zip_include(runtmp):
    # search zip, include on case sensitive match to name
    allzip = utils.get_test_data('prot/all.zip')

    runtmp.run_sourmash('sig', 'grep', '--dna', 'OS223', allzip)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'


def test_grep_5_zip_include_picklist(runtmp):
    # search zip, include on case sensitive match to name
    allzip = utils.get_test_data('prot/all.zip')

    pickfile = runtmp.output('pick.csv')
    with open(pickfile, 'w', newline="") as fp:
        w = csv.DictWriter(fp, fieldnames=['md5'])
        w.writeheader()
        w.writerow(dict(md5='09a08691ce52952152f0e866a59f6261'))
        w.writerow(dict(md5='38729c6374925585db28916b82a6f513'))

    runtmp.run_sourmash('sig', 'grep', '--dna', 'OS223', allzip,
                        '--picklist', f"{pickfile}:md5:md5")

    out = runtmp.last_result.out
    print(out)
    err = runtmp.last_result.err
    print(err)
    assert 'for given picklist, found 2 matches to 2 distinct values' in err

    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'


def test_grep_5_zip_include_case_insensitive(runtmp):
    # search zip, include on case insensitive match to name
    allzip = utils.get_test_data('prot/all.zip')

    runtmp.run_sourmash('sig', 'grep', '--dna', '-i', 'os223', allzip)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'


def test_grep_5_zip_exclude(runtmp):
    # search zip, exclude on case-sensitive match
    allzip = utils.get_test_data('prot/all.zip')

    runtmp.run_sourmash('sig', 'grep', '--dna', '-v', 'OS185', allzip)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'


def test_grep_5_zip_exclude_case_insensitive(runtmp):
    # search zip, exclude on case-insensitive match
    allzip = utils.get_test_data('prot/all.zip')

    runtmp.run_sourmash('sig', 'grep', '--dna', '-vi', 'os185', allzip)

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'


def test_grep_6_zip_manifest_csv(runtmp):
    # do --csv and use result as picklist
    allzip = utils.get_test_data('prot/all.zip')

    runtmp.run_sourmash('sig', 'grep', '--dna', 'OS223', allzip,
                        '--csv', 'match.csv')

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'

    # now run cat with picklist
    runtmp.run_sourmash('sig', 'cat', allzip,
                        '--picklist', 'match.csv::manifest')

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'


def test_grep_6_zip_manifest_csv_gz(runtmp):
    # do --csv and use result as picklist
    allzip = utils.get_test_data('prot/all.zip')

    runtmp.run_sourmash('sig', 'grep', '--dna', 'OS223', allzip,
                        '--csv', 'match.csv.gz')

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'

    # check that match.csv.gz is a gzip file
    with gzip.open(runtmp.output('match.csv.gz'), 'rt', newline='') as fp:
        fp.read()

    # now run cat with picklist
    runtmp.run_sourmash('sig', 'cat', allzip,
                        '--picklist', 'match.csv.gz::manifest')

    out = runtmp.last_result.out
    ss = load_signatures(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert 'Shewanella baltica OS223' in ss.name
    assert ss.md5sum() == '38729c6374925585db28916b82a6f513'


def test_sig_grep_7_lca(runtmp):
    # extract 47 from an LCA database, with --no-require-manifest
    allzip = utils.get_test_data('lca/47+63.lca.json')
    sig47 = utils.get_test_data('47.fa.sig')

    runtmp.sourmash('sig', 'grep', "50a9274021e4", allzip,
                    '--no-require-manifest', '-o', 'matches.sig')

    match = sourmash.load_file_as_signatures(runtmp.output('matches.sig'))
    match = list(match)[0]

    ss47 = sourmash.load_file_as_signatures(sig47)
    ss47 = list(ss47)[0]

    ss47 = ss47.to_mutable()
    ss47.minhash = ss47.minhash.downsample(scaled=10000)

    assert ss47.minhash == match.minhash


def test_sig_grep_7_picklist_md5_lca_fail(runtmp):
    # extract 47 from an LCA database, using a picklist w/full md5 => fail
    allzip = utils.get_test_data('lca/47+63.lca.json')

    # select on any of these attributes
    row = dict(exactName='NC_009665.1 Shewanella baltica OS185, complete genome',
               md5full='50a9274021e43eda8b2e77f8fa60ae8e',
               md5short='50a9274021e43eda8b2e77f8fa60ae8e'[:8],
               fullIdent='NC_009665.1',
               nodotIdent='NC_009665')

    # make picklist
    picklist_csv = runtmp.output('pick.csv')
    with open(picklist_csv, 'w', newline='') as csvfp:
        w = csv.DictWriter(csvfp, fieldnames=row.keys())
        w.writeheader()
        w.writerow(row)

    picklist_arg = f"{picklist_csv}:md5full:md5"
    with pytest.raises(SourmashCommandFailed) as exc:
        runtmp.sourmash('sig', 'grep', '50a92740', allzip,
                        '--picklist', picklist_arg,
                        '--no-require-manifest')

    # this happens b/c the implementation of 'grep' uses picklists, and
    # LCA databases don't support multiple picklists.
    print(runtmp.last_result.err)
    assert "This input collection doesn't support 'grep' with picklists." in runtmp.last_result.err


def test_sig_grep_8_count(runtmp):
    zips = ['prot/all.zip',
            'prot/dayhoff.sbt.zip',
            'prot/dayhoff.zip',
            'prot/hp.sbt.zip',
            'prot/hp.zip',
            'prot/protein.sbt.zip',
            'prot/protein.zip']

    zip_src = [ utils.get_test_data(x) for x in zips ]

    os.mkdir(runtmp.output('prot'))
    for src, dest in zip(zip_src, zips):
        shutil.copyfile(src, runtmp.output(dest))
    
    runtmp.sourmash('sig', 'grep', '-c', '0015939', *zips)

    out = runtmp.last_result.out
    err = runtmp.last_result.err

    print(out)
    print(err)

    assert "(no signatures will be saved because of --silent/--count)." in err

    for line in """\
6 matches: prot/all.zip
2 matches: prot/dayhoff.sbt.zip
2 matches: prot/dayhoff.zip
2 matches: prot/hp.sbt.zip
2 matches: prot/hp.zip
2 matches: prot/protein.sbt.zip
2 matches: prot/protein.zip
""".splitlines():
        assert line.strip() in out



"""extract one or more signatures"""

usage="""

### `sourmash signature extract` - extract signatures from a collection

Extract the specified signature(s) from a collection of signatures.

For example,

sourmash signature extract *.sig -k 21 --dna -o extracted.sig

will extract all nucleotide signatures calculated at k=21 from all
.sig files in the current directory.

There are currently two other useful selectors for `extract`: you can specify
(part of) an md5sum, as output in the CSVs produced by `search` and `gather`;
and you can specify (part of) a name.

For example,

sourmash signature extract tests/test-data/*.fa.sig --md5 09a0869

will extract the signature from `47.fa.sig` which has an md5sum of
`09a08691ce52952152f0e866a59f6261`; and 

sourmash signature extract tests/test-data/*.fa.sig --name NC_009665

will extract the same signature, which has an accession number of
`NC_009665.1`.

#### Using picklists with `sourmash sig extract`

As of sourmash 4.2.0, `extract` also supports picklists, a feature by
which you can select signatures based on values in a CSV file. See
[the command line docs](https://sourmash.readthedocs.io/en/latest/command-line.html) for more information.

"""

from sourmash.cli.utils import (add_moltype_args, add_ksize_arg,
                                add_picklist_args, add_pattern_args)


def subparser(subparsers):
    subparser = subparsers.add_parser('extract', description=__doc__, usage=usage)
    subparser.add_argument('signatures', nargs='*')
    subparser.add_argument(
        '-q', '--quiet', action='store_true',
        help='suppress non-error output'
    )
    subparser.add_argument(
        '-o', '--output', metavar='FILE',
        help='output signature to this file (default stdout)',
        default='-',
    )
    subparser.add_argument(
        '--md5', default=None,
        help='select signatures whose md5 contains this substring'
    )
    subparser.add_argument(
        '--name', default=None,
        help='select signatures whose name contains this substring'
    )
    subparser.add_argument(
        '-f', '--force', action='store_true',
        help='try to load all files as signatures'
    )
    subparser.add_argument(
        '--from-file',
        help='a text file containing a list of files to load signatures from'
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_pattern_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash
    return sourmash.sig.__main__.extract(args)



"""
Tests for the picklist API.
"""
import pytest
import sourmash

import sourmash_tst_utils as utils
from sourmash.picklist import SignaturePicklist


def test_load_empty_picklist_fail():
    empty = utils.get_test_data('picklist/empty.csv')

    pl = SignaturePicklist('manifest')
    with pytest.raises(ValueError):
        pl.load(empty, 'foo', allow_empty=False)


def test_load_empty_picklist_allow():
    empty = utils.get_test_data('picklist/empty.csv')

    pl = SignaturePicklist('manifest')
    pl.load(empty, 'foo', allow_empty=True)



from glob import glob
import os
import argparse
from sourmash.logging import notify
from sourmash.sourmash_args import check_scaled_bounds, check_num_bounds


def add_moltype_args(parser):
    parser.add_argument(
        '--protein', dest='protein', action='store_true',
        help='choose a protein signature; by default, a nucleotide signature is used'
    )
    parser.add_argument(
        '--no-protein', dest='protein', action='store_false',
        help='do not choose a protein signature')
    parser.set_defaults(protein=False)

    parser.add_argument(
        '--dayhoff', dest='dayhoff', action='store_true',
        help='choose Dayhoff-encoded amino acid signatures'
    )
    parser.add_argument(
        '--no-dayhoff', dest='dayhoff', action='store_false',
        help='do not choose Dayhoff-encoded amino acid signatures')
    parser.set_defaults(dayhoff=False)

    parser.add_argument(
        '--hp', '--hydrophobic-polar', dest='hp', action='store_true',
        help='choose hydrophobic-polar-encoded amino acid signatures'
    )
    parser.add_argument(
        '--no-hp', '--no-hydrophobic-polar', dest='hp', action='store_false',
        help='do not choose hydrophobic-polar-encoded amino acid signatures')
    parser.set_defaults(hp=False)

    parser.add_argument(
        '--dna', '--rna', '--nucleotide', dest='dna', default=None, action='store_true',
        help='choose a nucleotide signature (default: True)')
    parser.add_argument(
        '--no-dna', '--no-rna', '--no-nucleotide', dest='dna', action='store_false',
        help='do not choose a nucleotide signature')
    parser.set_defaults(dna=None)


def add_construct_moltype_args(parser):
    add_moltype_args(parser)
    parser.set_defaults(dna=True)


def add_ksize_arg(parser, *, default=None):
    "Add -k/--ksize to argparse parsers, with specified default."
    if default:
        message = f"k-mer size to select; default={default}"
    else:
        message = f"k-mer size to select; no default."

    parser.add_argument(
        '-k', '--ksize', metavar='K', default=default, type=int,
        help=message,
    )

#https://stackoverflow.com/questions/55324449/how-to-specify-a-minimum-or-maximum-float-value-with-argparse#55410582
def range_limited_float_type(arg):
    """ Type function for argparse - a float within some predefined bounds """
    min_val = 0
    max_val = 1
    try:
        f = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("\n\tERROR: Must be a floating point number.")
    if f < min_val or f > max_val:
        raise argparse.ArgumentTypeError(f"\n\tERROR: Argument must be >{str(min_val)} and <{str(max_val)}.")
    return f


def add_tax_threshold_arg(parser, containment_default=0.1, ani_default=None):
    parser.add_argument(
        '--containment-threshold', default=containment_default, type=range_limited_float_type,
        help=f'minimum containment threshold for classification; default={containment_default}',
    )
    parser.add_argument(
        '--ani-threshold', '--aai-threshold', default=ani_default, type=range_limited_float_type,
        help=f'minimum ANI threshold (nucleotide gather) or AAI threshold (protein gather) for classification; default={ani_default}',
    )


def add_picklist_args(parser):
    parser.add_argument(
        '--picklist', default=None,
        help="select signatures based on a picklist, i.e. 'file.csv:colname:coltype'"
    )
    parser.add_argument(
        '--picklist-require-all', default=False, action='store_true',
        help="require that all picklist values be found or else fail"
    )


def add_pattern_args(parser):
    parser.add_argument(
        '--include-db-pattern',
        default=None,
        help='search only signatures that match this pattern in name, filename, or md5'
    )
    parser.add_argument(
        '--exclude-db-pattern',
        default=None,
        help='search only signatures that do not match this pattern in name, filename, or md5'
    )


def opfilter(path):
    return not path.startswith('__') and path not in ['utils']


def command_list(dirpath):
    paths = glob(os.path.join(dirpath, '*.py'))
    filenames = [os.path.basename(path) for path in paths]
    basenames = [os.path.splitext(path)[0] for path in filenames if not path.startswith('__')]
    basenames = filter(opfilter, basenames)
    return sorted(basenames)


def add_scaled_arg(parser, default=None):
    parser.add_argument(
        '--scaled', metavar='FLOAT', type=check_scaled_bounds,
        help='downsample to this scaled; value should be between 100 and 1e6'
    )


def add_num_arg(parser, default=0):
    parser.add_argument(
        '-n', '--num-hashes', '--num', metavar='N', type=check_num_bounds, default=default,
        help='num value should be between 50 and 50000'
    )


def check_rank(args):
    """ Check '--rank'/'--position'/'--lin-position' argument matches selected taxonomy."""
    standard_ranks =['strain', 'species', 'genus', 'family', 'order', 'class', 'phylum', 'superkingdom']
    if args.lins:
        if args.rank.isdigit(): 
            return str(args.rank)
        raise argparse.ArgumentTypeError(f"Invalid '--rank'/'--position' input: '{args.rank}'. '--lins' is specified. Rank must be an integer corresponding to a LIN position.")
    elif args.rank in standard_ranks:
        return args.rank
    else:
        raise argparse.ArgumentTypeError(f"Invalid '--rank'/'--position' input: '{args.rank}'. Please choose: 'strain', 'species', 'genus', 'family', 'order', 'class', 'phylum', 'superkingdom'")


def add_rank_arg(parser):
    parser.add_argument(
        '-r', '--rank',
        '--position', '--lin-position',
        help="For non-default output formats. Classify to this rank (tax genome) or summarize taxonomy at this rank and above (tax metagenome). \
              Note that the taxonomy CSV must contain lineage information at this rank, and that LIN positions start at 0. \
              Choices: 'strain', 'species', 'genus', 'family', 'order', 'class', 'phylum', 'superkingdom' or an integer LIN position"
    )


def check_tax_outputs(args, rank_required = ["krona"], incompatible_with_lins = None, use_lingroup_format=False):
    "Handle ouput format combinations"
    # check that rank is passed for formats requiring rank.
    if not args.rank:
        if any(x in rank_required for x in args.output_format):
            raise ValueError(f"Rank (--rank) is required for {', '.join(rank_required)} output formats.")

    if args.lins:
        # check for outputs incompatible with lins
        if incompatible_with_lins:
            if any(x in args.output_format for x in incompatible_with_lins):
                raise ValueError(f"The following outputs are incompatible with '--lins': : {', '.join(incompatible_with_lins)}")
        # check that lingroup file exists if needed
        if args.lingroup:
            if use_lingroup_format and "lingroup" not in args.output_format:
                args.output_format.append("lingroup")
        elif "lingroup" in args.output_format:
            raise ValueError(f"Must provide lingroup csv via '--lingroup' in order to output a lingroup report.")
    elif args.lingroup or "lingroup" in args.output_format:
        raise ValueError(f"Must enable LIN taxonomy via '--lins' in order to use lingroups.")

    # check that only one output format is specified if writing to stdout
    if len(args.output_format) > 1:
        if args.output_base == "-":
            raise ValueError(f"Writing to stdout is incompatible with multiple output formats {args.output_format}")
    elif not args.output_format:
        # change to "human" for 5.0
        args.output_format = ["csv_summary"]

    return args.output_format



import sourmash_tst_utils as utils

@utils.in_tempdir
def test_bug_803(c):
    # can we do a 'sourmash search' on an LCA database and a query with abundance?
    query = utils.get_test_data('track_abund/47.fa.sig')
    lca_db = utils.get_test_data('lca/47+63.lca.json')

    c.run_sourmash('search', query, lca_db, '--ignore-abundance')
    print(c)
    assert 'NC_009665.1 Shewanella baltica OS185, complete genome' in str(c)



"""
Tests for the 'sourmash signature fileinfo' command line.
"""
import shutil
import os
import glob
import json

import pytest

import sourmash_tst_utils as utils
from sourmash_tst_utils import SourmashCommandFailed

## command line tests


def test_fileinfo_1_sig(runtmp):
    # get basic info on a signature
    sig47 = utils.get_test_data('47.fa.sig')

    shutil.copyfile(sig47, runtmp.output('sig47.sig'))
    runtmp.run_sourmash('sig', 'fileinfo', 'sig47.sig')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = """\
path filetype: MultiIndex
location: sig47.sig
is database? no
has manifest? yes
num signatures: 1
total hashes: 5177
summary of sketches:
   1 sketches with DNA, k=31, scaled=1000             5177
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


def test_fileinfo_1_sig_summarize(runtmp):
    # get basic info on a signature with 'summarize' as alias for fileinfo
    sig47 = utils.get_test_data('47.fa.sig')

    shutil.copyfile(sig47, runtmp.output('sig47.sig'))
    runtmp.run_sourmash('sig', 'summarize', 'sig47.sig')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = """\
path filetype: MultiIndex
location: sig47.sig
is database? no
has manifest? yes
num signatures: 1
total hashes: 5177
summary of sketches:
   1 sketches with DNA, k=31, scaled=1000             5177
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


def test_fileinfo_1_sig_abund(runtmp):
    # get basic info on a signature with abundance
    sig47 = utils.get_test_data('track_abund/47.fa.sig')

    shutil.copyfile(sig47, runtmp.output('sig47.sig'))
    runtmp.run_sourmash('sig', 'fileinfo', 'sig47.sig')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = """\
path filetype: MultiIndex
location: sig47.sig
is database? no
has manifest? yes
num signatures: 1
5177 total hashes
summary of sketches:
   1 sketches with DNA, k=31, scaled=1000, abund      5177 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


def test_fileinfo_2_lca(runtmp):
    # get basic info on an LCA database
    prot = utils.get_test_data('prot/protein.lca.json.gz')

    shutil.copyfile(prot, runtmp.output('protein.lca.json.gz'))
    runtmp.run_sourmash('sig', 'fileinfo', 'protein.lca.json.gz')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = """\
path filetype: LCA_Database
location: protein.lca.json.gz
is database? yes
has manifest? no
num signatures: 2
total hashes: 8214
summary of sketches:
   2 sketches with protein, k=19, scaled=100          8214
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


def test_fileinfo_3_sbt_zip(runtmp):
    # test on an SBT.zip
    prot = utils.get_test_data('prot/protein.sbt.zip')

    shutil.copyfile(prot, runtmp.output('protein.sbt.zip'))
    runtmp.run_sourmash('sig', 'fileinfo', 'protein.sbt.zip')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = """\
path filetype: SBT
location: protein.sbt.zip
is database? yes
has manifest? yes
num signatures: 2
total hashes: 8214
summary of sketches:
   2 sketches with protein, k=19, scaled=100          8214 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out, line.strip()


def test_fileinfo_4_zip(runtmp):
    # test on a ZipFileLinearIndex
    prot = utils.get_test_data('prot/all.zip')

    shutil.copyfile(prot, runtmp.output('all.zip'))
    runtmp.run_sourmash('sig', 'fileinfo', 'all.zip')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    # 'location' will be fully resolved, ignore it for now
    expected_output = """\
path filetype: ZipFileLinearIndex
is database? yes
has manifest? yes
num signatures: 8
total hashes: 31758
summary of sketches:
   2 sketches with dayhoff, k=19, scaled=100          7945 total hashes
   2 sketches with hp, k=19, scaled=100               5184 total hashes
   2 sketches with protein, k=19, scaled=100          8214 total hashes
   2 sketches with DNA, k=31, scaled=1000             10415 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


def test_fileinfo_4_zip_json_out(runtmp):
    # check --json-out
    prot = utils.get_test_data('prot/all.zip')

    shutil.copyfile(prot, runtmp.output('all.zip'))
    runtmp.run_sourmash('sig', 'fileinfo', 'all.zip', '--json-out')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    # should succeed as loading as JSON, with correct info
    vals = json.loads(out)

    assert vals['has_manifest']
    assert vals['is_database']
    assert vals['num_sketches'] == 8
    assert vals['path_filetype'] == 'ZipFileLinearIndex'
    assert vals['total_hashes'] == 31758

    d1 = {'ksize': 19, 'moltype': 'dayhoff', 'scaled': 100, 'num': 0, 'abund': False, 'count': 2, 'n_hashes': 7945}
    d2 = {'ksize': 19, 'moltype': 'hp', 'scaled': 100, 'num': 0, 'abund': False, 'count': 2, 'n_hashes': 5184}
    d3 = {'ksize': 19, 'moltype': 'protein', 'scaled': 100, 'num': 0, 'abund': False, 'count': 2, 'n_hashes': 8214}
    d4 = {'ksize': 31, 'moltype': 'DNA', 'scaled': 1000, 'num': 0, 'abund': False, 'count': 2, 'n_hashes': 10415}

    assert d1 in vals['sketch_info']
    assert d2 in vals['sketch_info']
    assert d3 in vals['sketch_info']
    assert d4 in vals['sketch_info']
    assert len(vals['sketch_info']) == 4


def test_fileinfo_4_zip_rebuild(runtmp):
    # test --rebuild
    prot = utils.get_test_data('prot/all.zip')

    shutil.copyfile(prot, runtmp.output('all.zip'))
    runtmp.run_sourmash('sig', 'fileinfo', 'all.zip', '--rebuild')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    # 'location' will be fully resolved, ignore it for now
    # CTB: note we're missing one of the 8 in the rebuilt, dna-sig.noext,
    # because it is not automatically included unless you load the zipfile
    # with traverse. This is intentional.
    expected_output = """\
path filetype: ZipFileLinearIndex
is database? yes
has manifest? yes
num signatures: 8
total hashes: 26581
summary of sketches:
   2 sketches with dayhoff, k=19, scaled=100          7945 total hashes
   2 sketches with hp, k=19, scaled=100               5184 total hashes
   2 sketches with protein, k=19, scaled=100          8214 total hashes
   1 sketches with DNA, k=31, scaled=1000             5238 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


def test_fileinfo_5_dir(runtmp):
    # test on a directory
    sig47 = utils.get_test_data('47.fa.sig')

    os.mkdir(runtmp.output('subdir'))

    shutil.copyfile(sig47, runtmp.output('subdir/sig47.sig'))
    runtmp.run_sourmash('sig', 'fileinfo', 'subdir/')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = """\
path filetype: MultiIndex
location: subdir/
is database? no
has manifest? yes
num signatures: 1
total hashes: 5177
summary of sketches:
   1 sketches with DNA, k=31, scaled=1000             5177 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


def test_fileinfo_6_pathlist(runtmp):
    # test on a pathlist
    sig47 = utils.get_test_data('47.fa.sig')
    shutil.copyfile(sig47, runtmp.output("47.fa.sig"))

    with open(runtmp.output('pathlist.txt'), 'wt') as fp:
        fp.write("47.fa.sig\n")

    runtmp.run_sourmash('sig', 'fileinfo', 'pathlist.txt')

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = """\
path filetype: MultiIndex
location: pathlist.txt
is database? no
has manifest? yes
num signatures: 1
total hashes: 5177
summary of sketches:
   1 sketches with DNA, k=31, scaled=1000             5177 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out


@pytest.mark.parametrize("db", ['v6.sbt.json', 'v5.sbt.json', 'v4.sbt.json',
                                'v3.sbt.json', 'v2.sbt.json', 'v1.sbt.json'])
def test_fileinfo_7_sbt_json(runtmp, db):
    # test on multiple versions of SBT JSON files
    dbfile = utils.get_test_data(db)

    runtmp.run_sourmash('sig', 'fileinfo', dbfile)

    out = runtmp.last_result.out
    print(runtmp.last_result.out)

    expected_output = f"""\
path filetype: SBT
location: {dbfile}
is database? yes
has manifest? no
num signatures: 7
total hashes: 3500
summary of sketches:
   7 sketches with DNA, k=31, num=500                 3500 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out, line.strip()


def test_sig_fileinfo_stdin(runtmp):
    # test on stdin
    sig = utils.get_test_data('prot/protein/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig')
    with open(sig, 'rt') as fp:
        data = fp.read()

    runtmp.run_sourmash('sig', 'fileinfo', '-', stdin_data=data)

    out = runtmp.last_result.out
    print(out)

    expected_output = """\
path filetype: MultiIndex
location: -
is database? no
has manifest? yes
num signatures: 1
total hashes: 3409
summary of sketches:
   1 sketches with protein, k=19, scaled=100          3409 total hashes
""".splitlines()
    for line in expected_output:
        assert line.strip() in out, line.strip()


def test_sig_fileinfo_does_not_exist(runtmp):
    # test on file that does not exist
    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash('sig', 'fileinfo', 'does-not-exist')

    assert "Cannot open 'does-not-exist' as a sourmash signature collection" in runtmp.last_result.err


def test_sig_fileinfo_8_manifest_works(runtmp):
    # test on a manifest with relative paths, in proper location
    mf = utils.get_test_data('scaled/mf.csv')
    runtmp.sourmash('sig', 'fileinfo', mf)

    out = runtmp.last_result.out
    print(out)

    assert '15 sketches with DNA, k=31, scaled=10000           717 total hashes' in out
    assert 'num signatures: 15' in out
    assert 'has manifest? yes' in out
    assert 'is database? yes' in out
    assert 'path filetype: StandaloneManifestIndex' in out


def test_sig_fileinfo_8_manifest_works_when_moved(runtmp):
    # test on a manifest with relative paths, when in wrong place
    # note: this works, unlike 'describe', because all the necessary info
    # for 'fileinfo' is in the manifest.
    mf = utils.get_test_data('scaled/mf.csv')
    shutil.copyfile(mf, runtmp.output('mf.csv'))

    runtmp.sourmash('sig', 'fileinfo', 'mf.csv')

    out = runtmp.last_result.out
    print(out)

    assert '15 sketches with DNA, k=31, scaled=10000           717 total hashes' in out
    assert 'num signatures: 15' in out
    assert 'has manifest? yes' in out
    assert 'is database? yes' in out
    assert 'path filetype: StandaloneManifestIndex' in out


def test_sig_fileinfo_9_sqldb_make(runtmp):
    # make a sqldb and run fileinfo on it
    gcf_all = glob.glob(utils.get_test_data('gather/GCF*.sig'))
    sqldb = runtmp.output('some.sqldb')

    runtmp.sourmash('sig', 'cat', '-k', '31', *gcf_all, '-o', sqldb)

    runtmp.sourmash('sig', 'fileinfo', sqldb)

    err = runtmp.last_result.err
    print(err)

    out = runtmp.last_result.out
    print(out)

    assert "12 sketches with DNA, k=31, scaled=10000           4540 total hashes" in out


def test_sig_fileinfo_9_sqldb_exists(runtmp):
    # run fileinfo on existing sqldb
    sqldb = utils.get_test_data('sqlite/index.sqldb')
    runtmp.sourmash('sig', 'fileinfo', sqldb)

    err = runtmp.last_result.err
    print(err)

    out = runtmp.last_result.out
    print(out)

    assert "path filetype: SqliteIndex" in out
    assert "2 sketches with DNA, k=31, scaled=1000             10415 total hashes" in out


def test_sig_fileinfo_9_sql_manifest(runtmp):
    # run fileinfo on existing sqldb
    sqldb = utils.get_test_data('sqlite/prot.sqlmf')
    runtmp.sourmash('sig', 'fileinfo', sqldb)

    err = runtmp.last_result.err
    print(err)

    out = runtmp.last_result.out
    print(out)

    assert "path filetype: StandaloneManifestIndex" in out
    assert "num signatures: 7" in out
    assert "1 sketches with DNA, k=31, scaled=1000             5238 total hashes" in out
    assert "2 sketches with hp, k=19, scaled=100               5184 total hashes" in out
    assert "2 sketches with dayhoff, k=19, scaled=100          7945 total hashes" in out
    assert "2 sketches with protein, k=19, scaled=100          8214 total hashes" in out


def test_sig_fileinfo_9_sql_lca_db(runtmp):
    # run fileinfo on existing sqldb
    sqldb = utils.get_test_data('sqlite/lca.sqldb')
    runtmp.sourmash('sig', 'fileinfo', sqldb)

    err = runtmp.last_result.err
    print(err)

    out = runtmp.last_result.out
    print(out)

    assert "path filetype: LCA_SqliteDatabase" in out
    assert "2 sketches with DNA, k=31, scaled=10000            1431 total hashes" in out



"""
Tests for 'sourmash sig collect'
"""
import pytest
import shutil
import os.path
import gzip

import sourmash
from sourmash.manifest import BaseCollectionManifest

import sourmash_tst_utils as utils
from sourmash_tst_utils import SourmashCommandFailed


def test_sig_collect_0_nothing(runtmp, manifest_db_format):
    # run with just output
    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'
    if manifest_db_format != 'sql': return

    runtmp.sourmash('sig', 'collect', '-o', f'mf.{ext}',
                    '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 0


def test_sig_collect_1_zipfile(runtmp, manifest_db_format):
    # collect a manifest from a .zip file
    protzip = utils.get_test_data('prot/protein.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', protzip, '-o', f'mf.{ext}',
                    '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list


def test_sig_collect_1_zipfile_csv_gz(runtmp):
    # collect a manifest from a .zip file, save to csv.gz
    protzip = utils.get_test_data('prot/protein.zip')

    runtmp.sourmash('sig', 'collect', protzip, '-o', 'mf.csv.gz',
                    '-F', 'csv')

    manifest_fn = runtmp.output('mf.csv.gz')

    # gzip, yes?
    print('XXX', manifest_fn)
    with gzip.open(manifest_fn, 'rt', newline='') as fp:
        fp.read()

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list


def test_sig_collect_1_zipfile_csv_gz_roundtrip(runtmp):
    # collect a manifest from a .zip file, save to csv.gz; then load again
    protzip = utils.get_test_data('prot/protein.zip')

    runtmp.sourmash('sig', 'collect', protzip, '-o', 'mf.csv.gz',
                    '-F', 'csv')

    manifest_fn = runtmp.output('mf.csv.gz')

    # gzip, yes?
    print('XXX', manifest_fn)
    with gzip.open(manifest_fn, 'rt', newline='') as fp:
        fp.read()

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list

    # can we read a csv.gz?
    runtmp.sourmash('sig', 'collect', 'mf.csv.gz', '-o', 'mf2.csv',
                    '-F', 'csv')

    manifest_fn2 = runtmp.output('mf2.csv')
    manifest2 = BaseCollectionManifest.load_from_filename(manifest_fn2)

    assert len(manifest2) == 2
    md5_list = [ row['md5'] for row in manifest2.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list



def test_sig_collect_2_exists_fail(runtmp, manifest_db_format):
    # collect a manifest from two .zip files
    protzip = utils.get_test_data('prot/protein.zip')
    allzip = utils.get_test_data('prot/protein.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', protzip, '-o', f'mf.{ext}',
                    '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list

    # now run with same filename - should fail
    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash('sig', 'collect', allzip, '-o', manifest_fn,
                        '-F', manifest_db_format)


def test_sig_collect_2_exists_merge(runtmp, manifest_db_format):
    # collect a manifest from two .zip files
    protzip = utils.get_test_data('prot/protein.zip')
    allzip = utils.get_test_data('prot/all.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', protzip, '-o', f'mf.{ext}',
                    '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list

    # now run with same filename - should merge
    runtmp.sourmash('sig', 'collect', allzip, '-o', manifest_fn,
                    '-F', manifest_db_format, '--merge')

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)
    assert len(manifest) == 10


def test_sig_collect_2_exists_sql_merge_csv(runtmp, manifest_db_format):
    # try to merge csv into sql
    protzip = utils.get_test_data('prot/protein.zip')
    allzip = utils.get_test_data('prot/all.zip')

    ext = 'sqlmf'

    # save as sql...
    runtmp.sourmash('sig', 'collect', protzip, '-o', f'mf.{ext}',
                    '-F', 'sql')

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list

    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash('sig', 'collect', allzip, '-o', manifest_fn,
                        '-F', 'csv', '--merge')

    assert "ERROR loading" in runtmp.last_result.err


def test_sig_collect_2_exists_csv_merge_sql(runtmp):
    # try to merge sql into csv
    protzip = utils.get_test_data('prot/protein.zip')
    allzip = utils.get_test_data('prot/all.zip')

    ext = 'csv'

    # save as csv...
    runtmp.sourmash('sig', 'collect', protzip, '-o', f'mf.{ext}',
                    '-F', 'csv')

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list

    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash('sig', 'collect', allzip, '-o', manifest_fn,
                        '-F', 'sql', '--merge')

    assert "ERROR loading" in runtmp.last_result.err


def test_sig_collect_2_no_exists_merge(runtmp, manifest_db_format):
    # test 'merge' when args.output doesn't already exist => warning
    protzip = utils.get_test_data('prot/protein.zip')
    allzip = utils.get_test_data('prot/all.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'
    manifest_fn = runtmp.output(f'mf.{ext}')

    # run with --merge but no previous:
    runtmp.sourmash('sig', 'collect', allzip, '-o', manifest_fn,
                    '-F', manifest_db_format, '--merge')

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)
    assert len(manifest) == 8

    err = runtmp.last_result.err
    print(err)
    assert "WARNING: --merge-previous specified, but output file" in err


def test_sig_collect_3_multiple(runtmp, manifest_db_format):
    # collect a manifest from two .zip files
    protzip = utils.get_test_data('prot/protein.zip')
    hpzip = utils.get_test_data('prot/hp.zip')
    dayzip = utils.get_test_data('prot/dayhoff.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', protzip, hpzip, dayzip,
                    '-o', f'mf.{ext}', '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 6
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list
    assert 'ea2a1ad233c2908529d124a330bcb672' in md5_list
    assert 'bb0e6d90df01b7bd5d0956a5f9e3ed12' in md5_list
    assert 'fbca5e5211e4d58427997fd5c8343e9a' in md5_list
    assert '1cbd888bf910f83ad8f1715509183223' in md5_list

    locations = set([ row['internal_location'] for row in manifest.rows ])
    assert protzip in locations
    assert hpzip in locations
    assert dayzip in locations
    assert len(locations) == 3, locations


def test_sig_collect_3_multiple_use_fromfile(runtmp, manifest_db_format):
    # collect a manifest from two .zip files using --from-file
    protzip = utils.get_test_data('prot/protein.zip')
    hpzip = utils.get_test_data('prot/hp.zip')
    dayzip = utils.get_test_data('prot/dayhoff.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    fromfile = runtmp.output('fromfile.txt')
    with open(fromfile, 'wt') as fp:
        print(protzip, file=fp)
        print(hpzip, file=fp)
        print(dayzip, file=fp)

    runtmp.sourmash('sig', 'collect', '--from-file', 'fromfile.txt',
                    '-o', f'mf.{ext}', '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 6
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '16869d2c8a1d29d1c8e56f5c561e585e' in md5_list
    assert '120d311cc785cc9d0df9dc0646b2b857' in md5_list
    assert 'ea2a1ad233c2908529d124a330bcb672' in md5_list
    assert 'bb0e6d90df01b7bd5d0956a5f9e3ed12' in md5_list
    assert 'fbca5e5211e4d58427997fd5c8343e9a' in md5_list
    assert '1cbd888bf910f83ad8f1715509183223' in md5_list

    locations = set([ row['internal_location'] for row in manifest.rows ])
    assert protzip in locations
    assert hpzip in locations
    assert dayzip in locations
    assert len(locations) == 3, locations


def test_sig_collect_4_multiple_from_sig(runtmp, manifest_db_format):
    # collect a manifest from sig files
    sig43 = utils.get_test_data('47.fa.sig')
    sig63 = utils.get_test_data('63.fa.sig')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', sig43, sig63,
                    '-o', f'mf.{ext}', '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '09a08691ce52952152f0e866a59f6261' in md5_list
    assert '38729c6374925585db28916b82a6f513' in md5_list

    locations = set([ row['internal_location'] for row in manifest.rows ])
    assert sig43 in locations
    assert sig63 in locations
    assert len(locations) == 2, locations


def test_sig_collect_4_multiple_from_sig_abspath(runtmp, manifest_db_format):
    # collect a manifest from sig files, forcing abspath
    sig43 = utils.get_test_data('47.fa.sig')
    sig63 = utils.get_test_data('63.fa.sig')

    shutil.copyfile(sig43, runtmp.output('47.fa.sig'))
    shutil.copyfile(sig63, runtmp.output('63.fa.sig'))

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', '47.fa.sig', '63.fa.sig', '--abspath',
                    '-o', f'mf.{ext}', '-F', manifest_db_format)

    print(runtmp.last_result.out)
    print(runtmp.last_result.err)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '09a08691ce52952152f0e866a59f6261' in md5_list
    assert '38729c6374925585db28916b82a6f513' in md5_list

    locations = set([ row['internal_location'] for row in manifest.rows ])
    print(locations)
    assert len(locations) == 2, locations

    for xx in locations:
        assert xx.startswith('/')


def test_sig_collect_4_multiple_no_abspath(runtmp, manifest_db_format):
    # collect a manifest from sig files, no abspath
    sig43 = utils.get_test_data('47.fa.sig')
    sig63 = utils.get_test_data('63.fa.sig')

    # copy files to tmp, where they will not have full paths
    shutil.copyfile(sig43, runtmp.output('47.fa.sig'))
    shutil.copyfile(sig63, runtmp.output('63.fa.sig'))

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', '47.fa.sig', '63.fa.sig',
                    '-o', f'mf.{ext}', '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [ row['md5'] for row in manifest.rows ]
    assert '09a08691ce52952152f0e866a59f6261' in md5_list
    assert '38729c6374925585db28916b82a6f513' in md5_list

    locations = set([ row['internal_location'] for row in manifest.rows ])
    print(locations)
    assert len(locations) == 2, locations
    assert '47.fa.sig' in locations
    assert '63.fa.sig' in locations


def test_sig_collect_5_no_manifest_sbt_fail(runtmp, manifest_db_format):
    # collect a manifest from files that don't have one
    sbt_zip = utils.get_test_data('v6.sbt.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash('sig', 'collect', sbt_zip,
                        '-o', f'mf.{ext}', '-F', manifest_db_format)


def test_sig_collect_5_no_manifest_sbt_succeed(runtmp, manifest_db_format):
    # generate a manifest from files that don't have one when --no-require
    sbt_zip = utils.get_test_data('v6.sbt.zip')

    ext = 'sqlmf' if manifest_db_format == 'sql' else 'csv'

    runtmp.sourmash('sig', 'collect', sbt_zip, '--no-require-manifest',
                    '-o', f'mf.{ext}', '-F', manifest_db_format)

    manifest_fn = runtmp.output(f'mf.{ext}')
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 7
    locations = set([ row['internal_location'] for row in manifest.rows ])
    assert len(locations) == 1, locations
    assert sbt_zip in locations



"""
Tests for the 'Index' class and protocol. All Index classes should support
this functionality.
"""

import pytest
import glob

import sourmash
from sourmash import SourmashSignature
from sourmash.index import (LinearIndex, ZipFileLinearIndex,
                            LazyLinearIndex, MultiIndex,
                            StandaloneManifestIndex,
                            IndexSearchResult)
from sourmash.index import CounterGather
from sourmash.index.sqlite_index import SqliteIndex
from sourmash.index.revindex import RevIndex
from sourmash.sbt import SBT, GraphFactory
from sourmash.manifest import CollectionManifest, BaseCollectionManifest
from sourmash.lca.lca_db import LCA_Database, load_single_database
from sourmash.minhash import (flatten_and_intersect_scaled,
                              flatten_and_downsample_scaled)

import sourmash_tst_utils as utils


def _load_three_sigs():
    # utility function - load & return these three sigs.
    sig2 = utils.get_test_data('2.fa.sig')
    sig47 = utils.get_test_data('47.fa.sig')
    sig63 = utils.get_test_data('63.fa.sig')

    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    ss47 = sourmash.load_one_signature(sig47)
    ss63 = sourmash.load_one_signature(sig63)

    return [ss2, ss47, ss63]


def build_linear_index(runtmp):
    ss2, ss47, ss63 = _load_three_sigs()

    lidx = LinearIndex()
    lidx.insert(ss2)
    lidx.insert(ss47)
    lidx.insert(ss63)

    return lidx


def build_lazy_linear_index(runtmp):
    lidx = build_linear_index(runtmp)
    return LazyLinearIndex(lidx)


def build_sbt_index(runtmp):
    ss2, ss47, ss63 = _load_three_sigs()
    
    factory = GraphFactory(5, 100, 3)
    root = SBT(factory, d=2)

    root.insert(ss2)
    root.insert(ss47)
    root.insert(ss63)

    return root


def build_sbt_index_save_load(runtmp):
    root = build_sbt_index(runtmp)
    out = runtmp.output('xyz.sbt.zip')
    root.save(out)

    return sourmash.load_file_as_index(out)


def build_zipfile_index(runtmp):
    from sourmash.save_load import SaveSignatures_ZipFile

    location = runtmp.output('index.zip')
    with SaveSignatures_ZipFile(location) as save_sigs:
        for ss in _load_three_sigs():
            save_sigs.add(ss)

    idx = ZipFileLinearIndex.load(location)
    return idx


def build_multi_index(runtmp):
    siglist = _load_three_sigs()
    lidx = LinearIndex(siglist)

    mi = MultiIndex.load([lidx], [None], None)
    return mi


def build_standalone_manifest_index(runtmp):
    sig2 = utils.get_test_data('2.fa.sig')
    sig47 = utils.get_test_data('47.fa.sig')
    sig63 = utils.get_test_data('63.fa.sig')

    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    ss47 = sourmash.load_one_signature(sig47)
    ss63 = sourmash.load_one_signature(sig63)

    siglist = [(ss2, sig2), (ss47, sig47), (ss63, sig63)]

    rows = []
    rows.extend((CollectionManifest.make_manifest_row(ss, loc) for ss, loc in siglist ))
    mf = CollectionManifest(rows)
    mf_filename = runtmp.output("mf.csv")
    
    mf.write_to_filename(mf_filename)

    idx = StandaloneManifestIndex.load(mf_filename)
    return idx


def build_lca_index(runtmp):
    siglist = _load_three_sigs()
    db = LCA_Database(31, 1000, 'DNA')
    for ss in siglist:
        db.insert(ss)

    return db


def build_lca_index_save_load(runtmp):
    db = build_lca_index(runtmp)
    outfile = runtmp.output('db.lca.json')
    db.save(outfile)

    return sourmash.load_file_as_index(outfile)


def build_sqlite_index(runtmp):
    filename = runtmp.output('idx.sqldb')
    db = SqliteIndex.create(filename)

    siglist = _load_three_sigs()
    for ss in siglist:
        db.insert(ss)

    return db


def build_revindex(runtmp):
    ss2, ss47, ss63 = _load_three_sigs()

    lidx = RevIndex(template=ss2.minhash)
    lidx.insert(ss2)
    lidx.insert(ss47)
    lidx.insert(ss63)

    return lidx


def build_lca_index_save_load_sql(runtmp):
    db = build_lca_index(runtmp)
    outfile = runtmp.output('db.lca.json')
    db.save(outfile, format='sql')

    x = load_single_database(outfile)
    db_load = x[0]

    return db_load


#
# create a fixture 'index_obj' that is parameterized by all of these
# building functions.
#

@pytest.fixture(params=[build_linear_index,
                        build_lazy_linear_index,
                        build_sbt_index,
                        build_zipfile_index,
                        build_multi_index,
                        build_standalone_manifest_index,
                        build_lca_index,
                        build_sbt_index_save_load,
                        build_lca_index_save_load,
                        build_sqlite_index,
                        build_lca_index_save_load_sql,
#                        build_revindex,
                        ]
)
def index_obj(request, runtmp):
    build_fn = request.param

    # build on demand
    return build_fn(runtmp)


###
### generic Index tests go here
###


def test_index_search_exact_match(index_obj):
    # search for an exact match
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss2, threshold=1.0)
    print([s[1].name for s in sr])
    assert len(sr) == 1
    assert sr[0].signature.minhash == ss2.minhash
    assert sr[0].score == 1.0


def test_index_search_lower_threshold(index_obj):
    # search at a lower threshold/multiple results with ss47
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss47, threshold=0.1)
    print([s[1].name for s in sr])
    assert len(sr) == 2
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss47.minhash
    assert sr[0].score == 1.0
    assert sr[1].signature.minhash == ss63.minhash
    assert round(sr[1].score, 2) == 0.32


def test_index_search_lower_threshold_2(index_obj):
    # search at a lower threshold/multiple results with ss63
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss63, threshold=0.1)
    print([s[1].name for s in sr])
    assert len(sr) == 2
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss63.minhash
    assert sr[0].score == 1.0
    assert sr[1].signature.minhash == ss47.minhash
    assert round(sr[1].score, 2) == 0.32


def test_index_search_higher_threshold_2(index_obj):
    # search at a higher threshold/one match
    ss2, ss47, ss63 = _load_three_sigs()

    # search for sig63 with high threshold => 1 match
    sr = index_obj.search(ss63, threshold=0.8)
    print([s[1].name for s in sr])
    assert len(sr) == 1
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss63.minhash
    assert sr[0].score == 1.0


def test_index_search_containment(index_obj):
    # search for containment at a low threshold/multiple results with ss63
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss63, do_containment=True, threshold=0.1)
    print([s[1].name for s in sr])
    assert len(sr) == 2
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss63.minhash
    assert sr[0].score == 1.0
    assert sr[1].signature.minhash == ss47.minhash
    assert round(sr[1].score, 2) == 0.48


def test_index_signatures(index_obj):
    # signatures works?
    siglist = list(index_obj.signatures())

    ss2, ss47, ss63 = _load_three_sigs()
    assert len(siglist) == 3

    # check md5sums, since 'in' doesn't always work
    md5s = set(( ss.md5sum() for ss in siglist ))
    assert ss2.md5sum() in md5s
    assert ss47.md5sum() in md5s
    assert ss63.md5sum() in md5s


def test_index_signatures_with_location(index_obj):
    # signatures_with_location works?
    siglist = list(index_obj.signatures_with_location())

    ss2, ss47, ss63 = _load_three_sigs()
    assert len(siglist) == 3

    # check md5sums, since 'in' doesn't always work
    md5s = set(( ss.md5sum() for ss, loc in siglist ))
    assert ss2.md5sum() in md5s
    assert ss47.md5sum() in md5s
    assert ss63.md5sum() in md5s


def test_index_len(index_obj):
    # len works?
    assert len(index_obj) == 3


def test_index_bool(index_obj):
    # bool works?
    assert bool(index_obj)


def test_index_location(index_obj):
    # location works?
    assert str(index_obj.location)


def test_index_manifest(index_obj):
    # manifest is either None or a BaseCollectionManifest
    manifest = index_obj.manifest
    if manifest is not None:
        assert isinstance(manifest, BaseCollectionManifest)


def test_index_select_basic(index_obj):
    # select does the basic thing ok
    idx = index_obj.select(ksize=31, moltype='DNA', abund=False,
                           containment=True, scaled=1000, num=0, picklist=None)

    assert len(idx) == 3
    siglist = list(idx.signatures())
    assert len(siglist) == 3

    # check md5sums, since 'in' doesn't always work
    md5s = set(( ss.md5sum() for ss in siglist ))
    ss2, ss47, ss63 = _load_three_sigs()
    assert ss2.md5sum() in md5s
    assert ss47.md5sum() in md5s
    assert ss63.md5sum() in md5s


def test_index_select_nada(index_obj):
    # select works ok when nothing matches!

    # CTB: currently this EITHER raises a ValueError OR returns an empty
    # Index object, depending on implementation. :think:
    # See: https://github.com/sourmash-bio/sourmash/issues/1940
    try:
        idx = index_obj.select(ksize=21)
    except ValueError:
        idx = LinearIndex([])

    assert len(idx) == 0
    siglist = list(idx.signatures())
    assert len(siglist) == 0


def test_index_prefetch(index_obj):
    # test basic prefetch
    ss2, ss47, ss63 = _load_three_sigs()

    # search for ss2
    results = []
    for result in index_obj.prefetch(ss2, threshold_bp=0):
        results.append(result)

    assert len(results) == 1
    assert results[0].signature.minhash == ss2.minhash

    # search for ss47 - expect two results
    results = []
    for result in index_obj.prefetch(ss47, threshold_bp=0):
        results.append(result)

    assert len(results) == 2
    assert results[0].signature.minhash == ss47.minhash
    assert results[1].signature.minhash == ss63.minhash


def test_index_best_containment(index_obj):
    # test basic containment search
    ss2, ss47, ss63 = _load_three_sigs()

    match = index_obj.best_containment(ss2)
    assert match
    assert match.score == 1.0
    assert match.signature.minhash == ss2.minhash

    match = index_obj.best_containment(ss47)
    assert match
    assert match.score == 1.0
    assert match.signature.minhash == ss47.minhash


def test_index_best_containment_threshold_1(index_obj):
    # test best_containment() method, in some detail
    ss2, ss47, ss63 = _load_three_sigs()

    # now construct query signatures with specific numbers of hashes --
    # note, these signatures all have scaled=1000.

    mins = list(sorted(ss2.minhash.hashes))
    new_mh = ss2.minhash.copy_and_clear()

    # query with empty hashes
    assert not new_mh
    with pytest.raises(ValueError):
        index_obj.best_containment(SourmashSignature(new_mh))

    # add one hash
    new_mh.add_hash(mins.pop())
    assert len(new_mh) == 1

    result = index_obj.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash

    # check with a threshold -> should be no results.
    with pytest.raises(ValueError):
        index_obj.best_containment(SourmashSignature(new_mh), threshold_bp=5000)

    # add three more hashes => length of 4
    new_mh.add_hash(mins.pop())
    new_mh.add_hash(mins.pop())
    new_mh.add_hash(mins.pop())
    assert len(new_mh) == 4

    result = index_obj.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash

    # check with a too-high threshold -> should be no results.
    with pytest.raises(ValueError):
        index_obj.best_containment(SourmashSignature(new_mh), threshold_bp=5000)


def test_best_containment_threshold_5(index_obj):
    # test gather() method, in some detail
    ss2, ss47, ss63 = _load_three_sigs()

    # now construct query signatures with specific numbers of hashes --
    # note, these signatures all have scaled=1000.

    mins = list(sorted(ss2.minhash.hashes.keys()))
    new_mh = ss2.minhash.copy_and_clear()

    # add five hashes
    for i in range(5):
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())

    # should get a result with no threshold (any match at all is returned)
    result = index_obj.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash

    # now, check with a threshold_bp that should be meet-able.
    result = index_obj.best_containment(SourmashSignature(new_mh), threshold_bp=5000)
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash


###
### CounterGather tests
###


def create_basic_counter_gather(runtmp):
    "Construct a CounterGather class."
    return CounterGather


class CounterGather_LinearIndex:
    """
    Provides an (inefficient) CounterGather-style class, for
    protocol testing purposes.
    """
    def __init__(self, orig_query):
        "Constructor - take a SourmashSignature that is the original query."
        orig_query_mh = orig_query.minhash
        if orig_query_mh.scaled == 0:
            raise ValueError

        # Index object used to actually track matches.
        self.idx = LinearIndex()
        self.orig_query_mh = orig_query_mh.copy().flatten()
        self.query_started = 0
        self.scaled = orig_query_mh.scaled
        self.locations = {}

    def add(self, ss, *, location=None, require_overlap=True):
        "Insert potential match."
        if self.query_started:
            raise ValueError("cannot add more signatures to counter after peek/consume")

        # skip duplicates
        md5 = ss.md5sum()
        if md5 in self.locations:
            return

        # confirm that this match has an overlap...
        add_mh = ss.minhash.flatten()
        overlap = self.orig_query_mh.count_common(add_mh, downsample=True)

        # ...figure out what scaled we are operating at now...
        if overlap:
            self.downsample(add_mh.scaled)
        elif require_overlap:
            raise ValueError("no overlap between query and signature!?")

        # ...and add to the Index, while also tracking location!
        self.idx.insert(ss)
        self.locations[md5] = location

    def signatures(self):
        "Yield all signatures"
        return self.idx.signatures()

    def downsample(self, scaled):
        "Track highest scaled across all possible matches."
        if scaled > self.scaled:
            self.scaled = scaled
        return self.scaled

    def peek(self, cur_query_mh, *, threshold_bp=0):
        """
        Find best match to current query within this CounterGather object.
        """
        self.query_started = 1

        scaled = self.downsample(cur_query_mh.scaled)
        cur_query_mh = flatten_and_downsample_scaled(cur_query_mh, scaled)

        # no hashes remaining? exit.
        if not self.orig_query_mh or not cur_query_mh:
            return []

        # verify current query is a subset of the original.
        if cur_query_mh.contained_by(self.orig_query_mh, downsample=True) < 1:
            raise ValueError("current query not a subset of original query")

        # did we get a match?
        res = self.idx.peek(cur_query_mh, threshold_bp=threshold_bp)
        if not res:
            return []
        sr, intersect_mh = res

        # got match - replace location & return.
        match = sr.signature
        md5 = match.md5sum()
        location = self.locations[md5]
        return IndexSearchResult(sr.score, match, location), intersect_mh

    def consume(self, *args, **kwargs):
        self.query_started = 1
        return self.idx.consume(*args, **kwargs)


class CounterGather_LCA:
    """
    Provides an alternative implementation of a CounterGather-style class,
    based on LCA_Database. This is currently just for protocol
    and API testing purposes.
    """
    def __init__(self, query):
        from sourmash.lca.lca_db import LCA_Database

        query_mh = query.minhash
        if query_mh.scaled == 0:
            raise ValueError("must use scaled MinHash")

        self.orig_query_mh = query_mh
        lca_db = LCA_Database(query_mh.ksize, query_mh.scaled,
                              query_mh.moltype)
        self.db = lca_db
        self.siglist = {}
        self.locations = {}
        self.query_started = 0

    def add(self, ss, *, location=None, require_overlap=True):
        "Add this signature into the counter."
        if self.query_started:
            raise ValueError("cannot add more signatures to counter after peek/consume")

        overlap = self.orig_query_mh.count_common(ss.minhash, True)
        if overlap:
            self.downsample(ss.minhash.scaled)
        elif require_overlap:
            raise ValueError("no overlap between query and signature!?")

        self.db.insert(ss)

        md5 = ss.md5sum()
        self.siglist[md5] = ss
        self.locations[md5] = location

    def signatures(self):
        "Yield all signatures."
        for ss in self.siglist.values():
            yield ss

    def downsample(self, scaled):
        "Track highest scaled across all possible matches."
        if scaled > self.db.scaled:
            self.db.downsample_scaled(scaled)
        return self.db.scaled

    def peek(self, query_mh, *, threshold_bp=0):
        "Return next possible match."
        from sourmash import SourmashSignature

        self.query_started = 1
        scaled = self.downsample(query_mh.scaled)
        query_mh = query_mh.downsample(scaled=scaled)

        if not self.orig_query_mh or not query_mh:
            return []

        if query_mh.contained_by(self.orig_query_mh, downsample=True) < 1:
            raise ValueError("current query not a subset of original query")

        query_ss = SourmashSignature(query_mh)

        # returns search_result, intersect_mh
        try:
            result = self.db.best_containment(query_ss, threshold_bp=threshold_bp)
        except ValueError:
            result = None

        if not result:
            return []

        cont = result.score
        match = result.signature

        intersect_mh = flatten_and_intersect_scaled(result.signature.minhash,
                                                    query_mh)

        md5 = result.signature.md5sum()
        location = self.locations[md5]

        new_sr = IndexSearchResult(cont, match, location)
        return [new_sr, intersect_mh]

    def consume(self, intersect_mh):
        self.query_started = 1


@pytest.fixture(params=[CounterGather,
                        CounterGather_LinearIndex,
                        CounterGather_LCA,
                        ]
)
def counter_gather_constructor(request):
    build_fn = request.param

    # build on demand
    return build_fn


def test_counter_get_signatures(counter_gather_constructor):
    # test .signatures() method
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(10, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(15, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    siglist = list(counter.signatures())
    assert len(siglist) == 3
    assert match_ss_1 in siglist
    assert match_ss_2 in siglist
    assert match_ss_3 in siglist


def _consume_all(query_mh, counter, threshold_bp=0):
    results = []
    query_mh = query_mh.to_mutable()

    last_intersect_size = None
    while 1:
        result = counter.peek(query_mh, threshold_bp=threshold_bp)
        if not result:
            break

        sr, intersect_mh = result
        print(sr.signature.name, len(intersect_mh))
        if last_intersect_size:
            assert len(intersect_mh) <= last_intersect_size

        last_intersect_size = len(intersect_mh)

        counter.consume(intersect_mh)
        query_mh.remove_many(intersect_mh.hashes)

        results.append((sr, len(intersect_mh)))

    return results


def test_counter_gather_1(counter_gather_constructor):
    # check a contrived set of non-overlapping gather results,
    # generated via CounterGather
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(10, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(15, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (['match1', 10],
                ['match2', 5],
                ['match3', 2],)
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_b(counter_gather_constructor):
    # check a contrived set of somewhat-overlapping gather results,
    # generated via CounterGather. Here the overlaps are structured
    # so that the gather results are the same as those in
    # test_counter_gather_1(), even though the overlaps themselves are
    # larger.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (['match1', 10],
                ['match2', 5],
                ['match3', 2],)
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_c_with_threshold(counter_gather_constructor):
    # check a contrived set of somewhat-overlapping gather results,
    # generated via CounterGather. Here the overlaps are structured
    # so that the gather results are the same as those in
    # test_counter_gather_1(), even though the overlaps themselves are
    # larger.
    # use a threshold, here.

    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter,
                           threshold_bp=3)

    expected = (['match1', 10],
                ['match2', 5])
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_d_diff_scaled(counter_gather_constructor):
    # test as above, but with different scaled.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear().downsample(scaled=10)
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear().downsample(scaled=20)
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear().downsample(scaled=30)
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (['match1', 10],
                ['match2', 5],
                ['match3', 2],)
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_d_diff_scaled_query(counter_gather_constructor):
    # test as above, but with different scaled for QUERY.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))

    match_mh_1 = query_mh.copy_and_clear().downsample(scaled=10)
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear().downsample(scaled=20)
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear().downsample(scaled=30)
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    # downsample query now -
    query_ss = SourmashSignature(query_mh.downsample(scaled=100), name='query')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (['match1', 10],
                ['match2', 5],
                ['match3', 2],)
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_e_abund_query(counter_gather_constructor):
    # test as above, but abund query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1, track_abundance=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear().flatten()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear().flatten()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear().flatten()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    # must flatten before peek!
    results = _consume_all(query_ss.minhash.flatten(), counter)

    expected = (['match1', 10],
                ['match2', 5],
                ['match3', 2],)
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_f_abund_match(counter_gather_constructor):
    # test as above, but abund query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1, track_abundance=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh.flatten(), name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name='match2')

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name='match3')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    # must flatten before peek!
    results = _consume_all(query_ss.minhash.flatten(), counter)

    expected = (['match1', 10],
                ['match2', 5],
                ['match3', 2],)
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_2(counter_gather_constructor):
    # check basic set of gather results on semi-real data,
    # generated via CounterGather
    testdata_combined = utils.get_test_data('gather/combined.sig')
    testdata_glob = utils.get_test_data('gather/GCF*.sig')
    testdata_sigs = glob.glob(testdata_glob)

    query_ss = sourmash.load_one_signature(testdata_combined, ksize=21)
    subject_sigs = [ (sourmash.load_one_signature(t, ksize=21), t)
                     for t in testdata_sigs ]

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    for ss, loc in subject_sigs:
        counter.add(ss, location=loc)

    results = _consume_all(query_ss.minhash, counter)

    expected = (['NC_003198.1', 487],
                ['NC_000853.1', 192],
                ['NC_011978.1', 169],
                ['NC_002163.1', 157],
                ['NC_003197.2', 152],
                ['NC_009486.1', 92],
                ['NC_006905.1', 76],
                ['NC_011080.1', 59],
                ['NC_011274.1', 42],
                ['NC_006511.1', 31],
                ['NC_011294.1', 7],
                ['NC_004631.1', 2])
    assert len(results) == len(expected)

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]
        print(sr_name, size)

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_exact_match(counter_gather_constructor):
    # query == match
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    # load up the counter; provide a location override, too.
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location='somewhere over the rainbow')

    results = _consume_all(query_ss.minhash, counter)
    assert len(results) == 1
    (sr, intersect_mh) = results[0]

    assert sr.score == 1.0
    assert sr.signature == query_ss
    assert sr.location == 'somewhere over the rainbow'


def test_counter_gather_multiple_identical_matches(counter_gather_constructor):
    # test multiple identical matches being inserted, with only one return
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    # create counter...
    counter = counter_gather_constructor(query_ss)

    # now add multiple identical matches.
    match_mh = query_mh.copy_and_clear()
    match_mh.add_many(range(5, 15))

    for name in 'match1', 'match2', 'match3':
        match_ss = SourmashSignature(match_mh, name=name)
        counter.add(match_ss, location=name)

    results = _consume_all(query_ss.minhash, counter)
    assert len(results) == 1

    sr, overlap_count = results[0]
    assert sr.score == 0.5
    assert overlap_count == 10

    # any one of the three is valid
    assert sr.location in ('match1', 'match2', 'match3')


def test_counter_gather_add_after_peek(counter_gather_constructor):
    # cannot add after peek or consume
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location='somewhere over the rainbow')

    counter.peek(query_ss.minhash)

    with pytest.raises(ValueError):
        counter.add(query_ss, location="try again")


def test_counter_gather_add_after_consume(counter_gather_constructor):
    # cannot add after peek or consume
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location='somewhere over the rainbow')

    counter.consume(query_ss.minhash)

    with pytest.raises(ValueError):
        counter.add(query_ss, location="try again")


def test_counter_gather_consume_empty_intersect(counter_gather_constructor):
    # check that consume works fine when there is an empty signature.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location='somewhere over the rainbow')

    # nothing really happens here :laugh:, just making sure there's no error
    counter.consume(query_ss.minhash.copy_and_clear())


def test_counter_gather_empty_initial_query(counter_gather_constructor):
    # check empty initial query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1, require_overlap=False)

    assert counter.peek(query_ss.minhash) == []


def test_counter_gather_num_query(counter_gather_constructor):
    # check num query
    query_mh = sourmash.MinHash(n=500, ksize=31)
    query_mh.add_many(range(0, 10))
    query_ss = SourmashSignature(query_mh, name='query')

    with pytest.raises(ValueError):
        counter_gather_constructor(query_ss)


def test_counter_gather_empty_cur_query(counter_gather_constructor):
    # test empty cur query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location='somewhere over the rainbow')

    cur_query_mh = query_ss.minhash.copy_and_clear()
    results = _consume_all(cur_query_mh, counter)
    assert results == []


def test_counter_gather_add_num_matchy(counter_gather_constructor):
    # test add num query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh = sourmash.MinHash(n=500, ksize=31)
    match_mh.add_many(range(0, 20))
    match_ss = SourmashSignature(match_mh, name='query')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    with pytest.raises(ValueError):
        counter.add(match_ss, location='somewhere over the rainbow')


def test_counter_gather_bad_cur_query(counter_gather_constructor):
    # test cur query that is not subset of original query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location='somewhere over the rainbow')

    cur_query_mh = query_ss.minhash.copy_and_clear()
    cur_query_mh.add_many(range(20, 30))
    with pytest.raises(ValueError):
        counter.peek(cur_query_mh)


def test_counter_gather_add_no_overlap(counter_gather_constructor):
    # check adding match with no overlap w/query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 10))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(10, 20))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    with pytest.raises(ValueError):
        counter.add(match_ss_1)

    assert counter.peek(query_ss.minhash) == []


def test_counter_gather_big_threshold(counter_gather_constructor):
    # check 'peek' with a huge threshold
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name='query')

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name='match1')

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)

    # impossible threshold:
    threshold_bp=30*query_ss.minhash.scaled
    results = counter.peek(query_ss.minhash, threshold_bp=threshold_bp)
    assert results == []


def test_counter_gather_empty_counter(counter_gather_constructor):
    # check empty counter
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_ss = SourmashSignature(query_mh, name='query')

    # empty counter!
    counter = counter_gather_constructor(query_ss)

    assert counter.peek(query_ss.minhash) == []
